#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fileencoding=utf-8

"""
Quartet Samping Tree Script
2017 James B. Pease
"""

import os
import sys
import argparse
import subprocess
from random import choice, randint, sample
from tree_reader import read_tree_string

PATTERN_CONVERT = {
    2: (6, 10, 18),
    4: (6, 12, 20),
    6: (14, 20),
    8: (10, 12, 24),
    10: (14, 26),
    12: (14, 28),
    14: (30),
    16: (18, 20, 24),
    18: (22, 26),
    20: (22, 28),
    22: (30),
    24: (26, 28),
    26: (30),
    28: (30),
    }


def run_ms(tree, params):
    """Runs the ms program to simulate sequence evolutions
        Arguments:
            params: dict from main args
    """
    nloci = params['nloci']
    theta = 4 * params['mu'] * params['popsize']
    if not params['quiet']:
        print("theta={}".format(theta))
    theta = theta * params['window']
    node_numbers = {}
    node_joins = []
    remaining_taxa = []
    i = 1
    for node in tree.iternodes(order="postorder"):
        if node.istip is True:
            node_numbers[node.label] = i
            node.label = i + 0
            remaining_taxa.append(node.label)
            i += 1
        else:
            node.label = node.children[randint(0, 1)].label
    for node in tree.iternodes(order="postorder"):
        if node.istip is False:
            valid_subnodes = [subnode.label for subnode in node.iternodes() if
                              subnode.istip is True and subnode.label in
                              remaining_taxa]
            print(valid_subnodes)
            node_order = sample(valid_subnodes, 2)
            print(node_order)
            node_joins.append(((round(max([x._calc_depth() for x in
                                           node.children]), 5) *
                                params['scale']) + params['addterm'],
                               node_order[0],
                               node_order[1]))
            remaining_taxa.remove(node_order[0])
    cmd = [params['ms_exec'],
           len(node_joins) + 1,
           int(nloci),
           "-t", theta,
           "-seeds", randint(100000, 999999),
           randint(100000, 999999), randint(100000, 999999),
           "-I", len(node_joins) + 1,
           " ".join(["1"]*(len(node_joins) + 1)),
           ]
    if params['recomb'] or params['rho']:
        if params['recomb']:
            rho = 4 * params['popsize'] * params['recomb'] * params['window']
        elif params['rho']:
            rho = params['window'] * params['rho']
        cmd.extend(["-r", str(rho), str(params['window'])])
        if not params['quiet']:
            print("rho={}".format(rho))
    cmd.extend(["-ej {} {} {}".format(x[0], x[1], x[2]) for x in
                sorted(node_joins)])
    if params['msource']:
        migration = params['mrate'] * 4 * params['popsize']
        cmd.extend(["-em", str(params['mtime_newer']), str(params['mdest']),
                    str(params['msource']), str(params['mrate'])])
        cmd.extend(["-eM", str(params['mtime_older']), '0'])
        if not params['quiet']:
            print("m={} {}>{} @{}-{}".format(
                migration, params['msource'], params['mdest'],
                params['mtime_newer'], params['mtime_older']))
    cmd.append("> {}.ms.tmp".format(params['out']))
    cmd = ' '.join([str(x) for x in cmd])
    if not params['quiet']:
        print(cmd)
    with open('{}.ms.log'.format(params['out']), 'w') as tmpfile:
        process = subprocess.Popen(
            cmd, shell=True, stdout=tmpfile, stderr=tmpfile)
        process.communicate()
    if not params['quiet']:
        print("ms done")
#    os.remove("seedms")
    for nname, nnum in sorted(node_numbers.items()):
        print(nname, nnum)
    return '{}.ms.tmp'.format(params['out'])


def run_raxml(alignfile, orig_nodes, params):
    if params['orig_tree'] is True:
        raxtreepath = params['infile']
    else:
        cmd = [params['raxml_exec'],
               "-s", alignfile,
               "-n", os.path.basename(alignfile) + ".simtree",
               "-m GTRGAMMA",
               "-p", randint(10000000, 9999999999),
               "-o", params['outgroup']]
        cmd = ' '.join([str(x) for x in cmd])
        if not params['quiet']:
            print(cmd)
        with open('{}.raxml.log'.format(params['out']), 'w') as tmpfile:
            process = subprocess.Popen(
                cmd, shell=True, stdout=tmpfile, stderr=tmpfile)
            process.communicate()
        raxtreepath = ("RAxML_bestTree." + os.path.basename(alignfile) +
                       ".simtree")
    raxraw = open(raxtreepath).read().strip()
    raxtree = read_tree_string(raxraw)
    for fname in os.listdir(os.curdir):
        if fname[0:6] == "RAxML_" and fname[-8:] == ".simtree":
            print(fname, "removed")
            os.remove(fname)
    i = 1
    orig_tree_matches = {}
    for node in raxtree.iternodes():
        if node.istip is False:
            rnode = 'RS' + str(i)
            orig_tree_matches[rnode] = None
            i += 1
            subnodes = [x.label for x in node.iternodes() if x.istip is True]
#            print(rnode, subnodes)
            match = False
            for onode in orig_nodes:
                if len(subnodes) > 0 and (
                        set(subnodes) == set(orig_nodes[onode][0])):
                    orig_tree_matches[rnode] = onode
                    print("MATCH", rnode, subnodes, "TO",
                          onode, orig_nodes[onode][0])
                    match = True
                    break
                elif len(subnodes) > 0 and (
                        set(subnodes) == set(orig_nodes[onode][1])):
                    orig_tree_matches[rnode] = onode
                    print("MATCH", rnode, subnodes, "TO",
                          onode, orig_nodes[onode][0])
                    match = True
                    break
            if match is False:
                print("NOMATCH", rnode, subnodes, )
    return orig_tree_matches, raxtree, raxraw


def run_qs(alignfile, treefile, params, rax_node_matches):
    qs_repdata = {}
    qs_notintree = []
    cmd = ["python3",
           params['qs_exec'],
           "--alignment", alignfile,
           "--tree", treefile,
           # "-L2",
           "-r SIM",
           "-N", params['qsrep'],
           "-T", params['qs_thread']]
    cmd = ' '.join([str(x) for x in cmd])
    if not params['quiet']:
        print(cmd)
    with open('{}.qs.log'.format(params['out']), 'w') as tmpfile:
        process = subprocess.Popen(
            cmd, shell=True, stdout=tmpfile, stderr=tmpfile)
        process.communicate()
    with open("SIM.node.scores.csv") as qsfile:
        firstline = True
        for line in qsfile:
            if firstline is True:
                firstline = False
                continue
            arr = line.rstrip().split(',')
            if arr[0][:2] == "QS":
                qdata = {'qc': arr[2], 'qd': arr[3], 'qi': arr[4]}
                xnodename = arr[0].replace("QS", "RS")
                if rax_node_matches[xnodename] is None:
                    qs_notintree.append(qdata)
                else:
                    qs_repdata[rax_node_matches[xnodename]] = qdata.copy()
    return qs_repdata, qs_notintree


def process_msfile(filepath, window):
    """Process a pre-computed ms output file
        Arguments:
            filepath: path of the msfile
            window: window length
    """
    aligns = []
    seqs = []
    pos = []
    with open(filepath, 'r') as tmpfile:
        tmpfile.readline()
        tmpfile.readline()
        for line in tmpfile:
            if '//' in line:
                if seqs:
                    aligns.append(dict([(pos[j], ''.join([
                        seqs[x][j] for x in range(len(seqs))]))
                                        for j in range(len(pos))]))
                seqs = []
            elif not line.strip():
                continue
            elif line[0] == 'p':
                pos = [int(float(x) * window)
                       for x in line.rstrip().split()[1:]]
            elif line[0] in 'sm':
                continue
            else:
                seqs.append(line.rstrip())
    if seqs:
        aligns.append(dict([(pos[j], ''.join(
            [seqs[x][j] for x in range(len(seqs))]))
             for j in range(len(pos))]))
    return aligns


def get_random_nuc():
    nuc_weights = ['A'] * 33 + ['T'] * 33 + ['G'] * 17 + ['C'] * 17
    return choice(nuc_weights)


def get_mut_pair(nuc):
    nuc_weights = {'A': 'GGCT', 'T': 'CCAG', 'C': 'TTAG', 'G': 'AACT'}
    return choice(nuc_weights[nuc])


def process_aligns(aligns, params):
    """Process alignments derived from an ms output file
        Arguments:
            aligns: list of alignment allele strings
            params: dict pass-through from main args
    """
    with open(params['out'], 'w') as outfile, open(
            params['out'] + '.parts', 'w') as partfile:
        coord_start = 0
        seqs = dict([(x + 1, []) for x in
                     range(len(list(aligns[0].values())[0]))])
        for i, align in enumerate(aligns):
            for position in range(params['window']):
                nuc = get_random_nuc()
                if position in align:
                    mutnuc = get_mut_pair(nuc)
                    for xpos, allele in enumerate(align[position]):
                        seqs[xpos + 1].append(mutnuc if allele == '1' else nuc)
                else:
                    for seq in seqs:
                        seqs[seq].append(nuc)
            partfile.write("DNA, Part{} = {}-{}\n".format(
                i, coord_start + 1, coord_start + params['window']))
            coord_start += params['window']
        outfile.write("{} {}\n".format(
            len(seqs), int(params['nloci'] * params['window'])))
        for seq in seqs:
            outfile.write("taxon_{}{}{}\n".format(
                seq, ' ' * (14 - len(str(seq))), ''.join(seqs[seq])))
    return ''


def generate_argparser():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('infile', type=os.path.abspath,
                        help="input tree file")
    parser.add_argument("-n", "--nrep", type=int,
                        default=100,
                        help=("scale tree by this factor"
                              "terminal branches"))
    parser.add_argument("-q", "--qsrep", type=int,
                        default=100,
                        help=("scale tree by this factor"
                              "terminal branches"))
    parser.add_argument("-s", "--scale", type=float,
                        default=1.,
                        help=("scale tree by this factor"
                              "terminal branches"))
    parser.add_argument("-a", "--addterm", type=float,
                        default=0.,
                        help=("add this length to"
                              "terminal branches"))
    parser.add_argument("-l", "--nloci", type=float,
                        default=1,
                        help=("number of reps for ms"))
    parser.add_argument('-o', '--out', type=os.path.abspath,
                        required=True, help="output file")
    parser.add_argument("-F", "--msfile",
                        help=("use pre-computed ms output"
                              "file instead of running ms."))
    parser.add_argument('-w', "--window", type=int, default=100000,
                        help="length (bp) of windows")
    parser.add_argument('-N', "--popsize", type=int, default=1e5,
                        help="Ne, effective population size (default=1e5)")
    parser.add_argument('-u', "--mu", type=float, default=1e-8,
                        help="""per site per generation mutation rate
                                (default=7e-9)""")
    parser.add_argument('-M', "--mrate", type=float, default=5e-4,
                        help="""per individual per generation migration
                                rate (default=5e-4)""")
    parser.add_argument('-S', "--msource", type=int,
                        help="1-based index of migration source population")
    parser.add_argument('-D', "--mdest", type=int,
                        help="1-based index of migration recipient population")
    parser.add_argument('-T', "--mtimes", type=float, nargs=2,
                        help="time bounds for the migration period")
    parser.add_argument('-r', "--recomb", type=float, default=0.,
                        help="""per site per generation recombination rate
                                (default=0)""")
    parser.add_argument("-R", "--rho", type=float,
                        help="""specific rho = 4*Ne*mu instead of using
                                --recomb""")
    parser.add_argument("--ms_exec", default="ms",
                        help="""path to ms executable""")
    parser.add_argument("--raxml_exec", default="raxml",
                        help="""path to raxml executable""")
    parser.add_argument("-t", "--qs-thread", type=int, default=1)
    parser.add_argument("-g", "--outgroup")
    parser.add_argument("--orig-tree", action="store_true",
                        help="evaluate original tree instead of RAxML")
    parser.add_argument("--qs_exec",
                        help="""path to quartetsampling.py""")
    parser.add_argument("--nconverge", type=int, default=0,
                        help="number of convergent sites per window")
    parser.add_argument("--quiet", action='store_true',
                        help='suppress screen output')
    return parser


def main(arguments=None):
    arguments = arguments if arguments is not None else sys.argv[1:]
    parser = generate_argparser()
    args = parser.parse_args(args=arguments)
    params = vars(args)
    qs_total = []
    qs_total_not = []
    params['mtime_newer'] = None if args.mtimes is None else max(args.mtimes)
    params['mtime_older'] = None if args.mtimes is None else min(args.mtimes)
    xfile = open(params['out'] + '.total.raxtrees', 'w')
    xfile.close()
    with open(args.infile) as treefile:
        for line in treefile:
            treestr = line.rstrip()
            orig_tree = read_tree_string(treestr)
            break
    all_tips = []
    orig_nodes = {}
    i = 1
    for node in orig_tree.iternodes():
        if node.istip is True:
            all_tips.append(node.label)
        else:
            nodename = 'QS' + str(i)
            i += 1
            orig_nodes[nodename] = [[], []]
            for subnode in node.iternodes():
                if subnode.istip is True:
                    orig_nodes[nodename][0].append(subnode.label)
    for xnode in orig_nodes:
        orig_nodes[xnode][0] = list(sorted(orig_nodes[xnode][0]))
        orig_nodes[xnode][1] = list(sorted([x for x in all_tips
                                            if x not in
                                            orig_nodes[xnode][0]]))
    for rep in range(args.nrep):
        mstmpfile = run_ms(orig_tree, params=params)
        aligns = process_msfile(mstmpfile, args.window)
        process_aligns(aligns, params=vars(args))
        rax_node_matches, raxtree, raxraw = run_raxml(
            params['out'], orig_nodes, params=params)
        with open(params['out'] + '.total.raxtrees', 'a') as outfile:
            outfile.write(raxraw + "\n")
        with open(params['out'] + '.raxtree', 'w') as outfile:
            outfile.write(raxraw + "\n")
        qs_repdata, qs_notintree = run_qs(
            params['out'], params['out'] + ".raxtree", params,
            rax_node_matches)
        qs_total_not.extend(qs_notintree)
        qs_total.append(qs_repdata)
        print(qs_repdata, qs_notintree)
    nodeheaders = list(sorted(orig_nodes.keys()))
    with open(params['out'] + ".total.qc", 'w') as qcfile, open(
        params['out'] + ".total.qd", 'w') as qdfile, open(
            params['out'] + ".total.qi", 'w') as qifile:
        qcfile.write(",".join(nodeheaders) + "\n")
        qdfile.write(",".join(nodeheaders) + "\n")
        qifile.write(",".join(nodeheaders) + "\n")
        for entry in qs_total:
            qcfile.write(",".join([entry[x].get('qc', 'NA')
                                   if x in entry else
                                   'NA' for x in nodeheaders]) + "\n")
            qdfile.write(",".join([entry[x].get('qd', 'NA')
                                   if x in entry else
                                   'NA' for x in nodeheaders]) + "\n")
            qifile.write(",".join([entry[x].get('qi', 'NA')
                                   if x in entry else
                                   'NA' for x in nodeheaders]) + "\n")
    with open(params['out'] + '.total.not', 'w') as qxfile:
        qxfile.write("qc,qd,qi\n")
        for entry in qs_total_not:
            qxfile.write(",".join([entry['qc'], entry['qd'], entry['qi']]) +
                         "\n")

    return ''


if __name__ == '__main__':
    main()
